@@ -9,7 +9,7 @@
 # License   : Revised BSD License
 # -----------------------------------------------------------------------------
 # Creation  : 26-Apr-2010
-# Last mod  : 29-Nov-2012
+# Last mod  : 21-Feb-2013
 # -----------------------------------------------------------------------------
 
 """
@@ -35,15 +35,15 @@ See also:
 - EC2, fabric, and "err: stdin: is not a tty"
   <http://blog.markfeeney.com/2009/12/ec2-fabric-and-err-stdin-is-not-tty.html>`_
 
-:copyright: (c) 2011,2012 by Sébastien Pierre.
+:copyright: (c) 2011-2013 by Sébastien Pierre.
 :license:   BSD, see LICENSE for more details.
 """
 
 from __future__ import with_statement
-import base64, zlib, hashlib, os, re, string, tempfile, subprocess, types, functools, StringIO
+import base64, hashlib, os, re, string, tempfile, subprocess, types, functools, StringIO
 import fabric, fabric.api, fabric.operations, fabric.context_managers
 
-VERSION         = "0.4.4"
+VERSION         = "0.5.0"
 RE_SPACES       = re.compile("[\s\t]+")
 MAC_EOL         = "\n"
 UNIX_EOL        = "\n"
@@ -53,16 +53,17 @@ MODE_SUDO       = "CUISINE_MODE_SUDO"
 SUDO_PASSWORD   = "CUISINE_SUDO_PASSWORD"
 OPTION_PACKAGE  = "CUISINE_OPTION_PACKAGE"
 OPTION_PYTHON_PACKAGE  = "CUISINE_OPTION_PYTHON_PACKAGE"
+
 AVAILABLE_OPTIONS = dict(
 	package=["apt", "yum", "zypper"],
 	python_package=["easy_install","pip"]
 )
+
 DEFAULT_OPTIONS = dict(
 	package="apt",
 	python_package="pip"
 )
 
-
 def sudo_password(password=None):
 	"""Sets the password for the sudo command."""
 	if password is None:
@@ -123,6 +124,11 @@ def is_local():  return mode(MODE_LOCAL)
 def is_remote(): return not mode(MODE_LOCAL)
 def is_sudo():   return mode(MODE_SUDO)
 
+def connect( host, user="root" ):
+	"""Initiates a connection with the given host and user"""
+	fabric.api.env.host_string = host
+	fabric.api.env.user        = user
+
 # =============================================================================
 #
 # OPTIONS
@@ -351,7 +357,7 @@ def file_local_read(location):
 def file_read(location):
 	"""Reads the *remote* file at the given location."""
 	# NOTE: We use base64 here to be sure to preserve the encoding (UNIX/DOC/MAC) of EOLs
-	return base64.b64decode(run('cat "%s" | base64' % (location)))
+	return base64.b64decode(run('cat "%s" | openssl base64' % (location)))
 
 def file_exists(location):
 	"""Tests if there is a *remote* file at the given location."""
@@ -395,11 +401,11 @@ def file_write(location, content, mode=None, owner=None, group=None, sudo=None,
 	# FIXME: Big files are never transferred properly!
 	# Gets the content signature and write it to a secure tempfile
 	use_sudo       = sudo if sudo is not None else is_sudo()
-	sig            = hashlib.sha256(content).hexdigest()
+	sig            = hashlib.md5(content).hexdigest()
 	fd, local_path = tempfile.mkstemp()
 	os.write(fd, content)
 	# Upload the content if necessary
-	if not file_exists(location) or sig != file_sha256(location):
+	if sig != file_md5(location):
 		if is_local():
 			with mode_sudo(use_sudo):
 				run('cp "%s" "%s"'%(local_path,location))
@@ -409,23 +415,24 @@ def file_write(location, content, mode=None, owner=None, group=None, sudo=None,
 			#fabric.operations.put(local_path, location, use_sudo=use_sudo)
 			# Hides the output, which is especially important
 			with fabric.context_managers.settings(
-				fabric.api.hide('warnings', 'running', 'stdout'),
+				#fabric.api.hide('warnings', 'running', 'stdout'),
 				warn_only=True,
 				**{MODE_SUDO: use_sudo}
 			):
 				# See: http://unix.stackexchange.com/questions/22834/how-to-uncompress-zlib-data-in-unix
-				with mode_sudo(use_sudo):
-					result = run("echo '%s' | base64 --decode | openssl zlib -d > \"%s\"" % (base64.b64encode(zlib.compress(content)), location))
+				result = run("echo '%s' | openssl base64 -A -d -out \"%s\"" % (base64.b64encode(content), location))
 				if result.failed:
 					fabric.api.abort('Encountered error writing the file %s: %s' % (location, result))
+					
 
 	# Remove the local temp file
+	os.fsync(fd)
 	os.close(fd)
 	os.unlink(local_path)
 	# Ensures that the signature matches
 	if check:
 		with mode_sudo(use_sudo):
-			file_sig = file_sha256(location)
+			file_sig = file_md5(location)
 		assert sig == file_sig, "File content does not matches file: %s, got %s, expects %s" % (location, repr(file_sig), repr(sig))
 	with mode_sudo(use_sudo):
 		file_attribs(location, mode=mode, owner=owner, group=group)
@@ -446,8 +453,8 @@ def file_upload(remote, local, sudo=None):
 	f       = file(local, 'rb')
 	content = f.read()
 	f.close()
-	sig     = hashlib.sha256(content).hexdigest()
-	if not file_exists(remote) or sig != file_sha256(remote):
+	sig     = hashlib.md5(content).hexdigest()
+	if not file_exists(remote) or sig != file_md5(remote):
 		if is_local():
 			if use_sudo:
 				sudo('cp "%s" "%s"'%(local,remote))
@@ -469,12 +476,12 @@ def file_update(location, updater=lambda x: x):
 	assert file_exists(location), "File does not exists: " + location
 	new_content = updater(file_read(location))
 	# assert type(new_content) in (str, unicode, fabric.operations._AttributeString), "Updater must be like (string)->string, got: %s() = %s" %  (updater, type(new_content))
-	run('echo "%s" | base64 -d > "%s"' % (base64.b64encode(new_content), location))
+	run('echo "%s" | openssl base64 -A -d -out "%s"' % (base64.b64encode(new_content), location))
 
 def file_append(location, content, mode=None, owner=None, group=None):
 	"""Appends the given content to the remote file at the given
 	location, optionally updating its mode/owner/group."""
-	run('echo "%s" | base64 -d >> "%s"' % (base64.b64encode(content), location))
+	run('echo "%s" | openssl base64 -A -d >> "%s"' % (base64.b64encode(content), location))
 	file_attribs(location, mode, owner, group)
 
 def file_unlink(path):
@@ -503,6 +510,14 @@ def file_sha256(location):
 	sig = run('shasum -a 256 "%s" | cut -d" " -f1' % (location)).split("\n")
 	return sig[-1].strip()
 
+def file_md5(location):
+	"""Returns the MD5 sum (as a hex string) for the remote file at the given location."""
+	# NOTE: In some cases, sudo can output errors in here -- but the errors will
+	# appear before the result, so we simply split and get the last line to
+	# be on the safe side.
+	sig = run('md5sum "%s" | cut -d" " -f1' % (location)).split("\n")
+	return sig[-1].strip()
+
 # =============================================================================
 #
 # PROCESS OPERATIONS
@@ -513,16 +528,26 @@ def process_find(name, exact=False):
 	"""Returns the pids of processes with the given name. If exact is `False`
 	it will return the list of all processes that start with the given
 	`name`."""
-	processes = run("ps aux | grep {0}".format(name))
-	res       = []
+	is_string = isinstance(name,str) or isinstance(name,unicode)
+	# NOTE: ps -A seems to be the only way to not have the grep appearing
+	# as well
+	if is_string: processes = run("ps -A | grep {0} ; true".format(name))
+	else:         processes = run("ps -A")
+	res = []
 	for line in processes.split("\n"):
 		if not line.strip(): continue
-		line = RE_SPACES.split(line,10)
+		line = RE_SPACES.split(line,3)
+		# 3010 pts/1    00:00:07 gunicorn
+		# PID  TTY      TIME     CMD
+		# 0    1        2        3
 		# We skip lines that are not like we expect them (sometimes error
 		# message creep up the output)
-		if len(line) < 11: continue
-		user, pid, cpu, mem, vsz, rss, tty, stat, start, time, command = line
-		if (exact and command == name) or (not exact and command.startswith(name)):
+		if len(line) < 4: continue
+		pid, tty, time, command = line
+		if is_string:
+			if pid and ((exact and command == name) or (not exact and command.find(name) >= 0)):
+				res.append(pid)
+		elif name(line) and pid:
 			res.append(pid)
 	return res
 
@@ -531,7 +556,7 @@ def process_kill(name, signal=9, exact=False):
 	it will return the list of all processes that start with the given
 	`name`."""
 	for pid in process_find(name, exact):
-		run("kill -s {0} {1}".format(signal, pid))
+		run("kill -s {0} {1} ; true".format(signal, pid))
 
 # =============================================================================
 #
@@ -573,7 +598,7 @@ def dir_ensure(location, recursive=False, mode=None, owner=None, group=None):
 # =============================================================================
 
 @dispatch
-def package_upgrade():
+def package_upgrade(distupgrade=False):
 	"""Updates every package present on the system."""
 
 @dispatch
@@ -596,12 +621,16 @@ def package_ensure(package, update=False):
 def package_clean(package=None):
 	"""Clean the repository for un-needed files."""
 
+@dispatch
+def package_remove(package, autoclean=False):
+	"""Remove package and optionally clean unused packages"""
+
 # -----------------------------------------------------------------------------
 # APT PACKAGE (DEBIAN/UBUNTU)
 # -----------------------------------------------------------------------------
 
 def repository_ensure_apt(repository):
-	sudo("add-apt-repository " + repository)
+	sudo("add-apt-repository --yes" + repository)
 
 def package_update_apt(package=None):
 	if package == None:
@@ -611,8 +640,11 @@ def package_update_apt(package=None):
 			package = " ".join(package)
 		sudo('DEBIAN_FRONTEND=noninteractive apt-get --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" upgrade ' + package)
 
-def package_upgrade_apt():
-	sudo('DEBIAN_FRONTEND=noninteractive apt-get --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" upgrade')
+def package_upgrade_apt(distupgrade=False):
+	if distupgrade:
+		sudo('DEBIAN_FRONTEND=noninteractive apt-get --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" dist-upgrade')
+	else:
+		sudo('DEBIAN_FRONTEND=noninteractive apt-get --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" upgrade')
 
 def package_install_apt(package, update=False):
 	if update:
@@ -622,16 +654,27 @@ def package_install_apt(package, update=False):
 	sudo("DEBIAN_FRONTEND=noninteractive apt-get --yes install %s" % (package))
 
 def package_ensure_apt(package, update=False):
-	status = run("dpkg-query -W -f='${Status}' %s ; true" % package)
-	if status.find("not-installed") != -1 or status.find("installed") == -1:
-		package_install(package, update)
+	"""Ensure apt packages are installed"""
+	if not isinstance(package, basestring):
+		package = " ".join(package)
+	status = run("dpkg-query -W -f='${Status} ' %s ; true" % package)
+	if ('No packages found' in status) or ('not-installed' in status) or ("installed" not in status):
+		package_install_apt(package)
 		return False
 	else:
-		if update: package_update(package)
+		if update:
+			package_update_apt(package)
 		return True
 
 def package_clean_apt(package=None):
-	pass
+	if type(package) in (list, tuple):
+		package = " ".join(package)
+	sudo("DEBIAN_FRONTEND=noninteractive apt-get -y --purge remove %s" % package)
+
+def package_remove_apt(package, autoclean=False):
+	sudo('DEBIAN_FRONTEND=noninteractive apt-get --yes -o Dpkg::Options::="--force-confdef" -o Dpkg::Options::="--force-confold" remove ' + package)
+	if autoclean:
+		sudo('apt-get --yes autoclean')
 
 # -----------------------------------------------------------------------------
 # YUM PACKAGE (RedHat, CentOS)
@@ -759,7 +802,7 @@ def python_package_upgrade_pip(package,E=None):
 	if E:
 		E='-E %s' %E
 	else:
-		E=''   
+		E=''
 	run('pip upgrade %s %s' %(E,package))
 
 def python_package_install_pip(package=None,r=None,pip=None):
@@ -779,7 +822,7 @@ def python_package_install_pip(package=None,r=None,pip=None):
 	else:
 		raise Exception("Either a package name or the requirements file has to be provided.")
 
-def python_package_ensure_pip(package=None,r=None, pip=None):
+def python_package_ensure_pip(package=None, r=None, pip=None):
 	'''
 	The "package" argument, defines the name of the package that will be ensured.
 	The argument "r" referes to the requirements file that will be used by pip and
@@ -806,7 +849,6 @@ def python_package_remove_pip(package, E=None, pip=None):
 	pip=pip or fabric.api.env.get('pip','pip')
 	return run('%s uninstall %s' %(pip,package))
 
-
 # -----------------------------------------------------------------------------
 # EASY_INSTALL PYTHON PACKAGE MANAGER
 # -----------------------------------------------------------------------------
@@ -866,14 +908,17 @@ def command_ensure(command, package=None):
 #
 # =============================================================================
 
-def user_passwd(name, passwd, encrypted_passwd=False):
+def user_passwd(name, passwd, encrypted_passwd=True):
 	"""Sets the given user password."""
 	encoded_password = base64.b64encode("%s:%s" % (name, passwd))
-	encryption = " -e" if encrypted_passwd else ""
-	sudo("echo %s | base64 --decode | chpasswd%s" % (encoded_password, encryption))
+	if encrypted_passwd:
+		sudo("usermod -p '%s' %s" % (passwd,name))
+	else:
+		sudo("echo %s | openssl base64 -A -d | chpasswd" % (encoded_password))
 
 def user_create(name, passwd=None, home=None, uid=None, gid=None, shell=None,
-				uid_min=None, uid_max=None, encrypted_passwd=False):
+				uid_min=None, uid_max=None, encrypted_passwd=True,
+				fullname=None):
 	"""Creates the user with the given name, optionally giving a
 	specific password/home/uid/gid/shell."""
 	options = ["-m"]
@@ -892,9 +937,11 @@ def user_create(name, passwd=None, home=None, uid=None, gid=None, shell=None,
 		options.append("-K UID_MIN='%s'" % (uid_min))
 	if uid_max:
 		options.append("-K UID_MAX='%s'" % (uid_max))
+	if fullname:
+		options.append("--gecos='%s'" % (fullname))
 	sudo("useradd %s '%s'" % (" ".join(options), name))
 	if passwd:
-		user_passwd(name,passwd,encrypted_passwd)
+		user_passwd(name=name,passwd=passwd,encrypted_passwd=encrypted_passwd)
 
 def user_check(name=None, uid=None):
 	"""Checks if there is a user defined with the given name,
@@ -912,7 +959,7 @@ def user_check(name=None, uid=None):
 	if d:
 		d = d.split(":")
 		assert len(d) >= 7, "/etc/passwd entry is expected to have at least 7 fields, got %s in: %s" % (len(d), ":".join(d))
-		results = dict(name=d[0], uid=d[2], gid=d[3], home=d[5], shell=d[6])
+		results = dict(name=d[0], uid=d[2], gid=d[3], fullname=d[4], home=d[5], shell=d[6])
 		s = sudo("cat /etc/shadow | egrep '^%s:' | awk -F':' '{print $2}'" % (results['name']))
 		if s: results['passwd'] = s
 	if results:
@@ -920,7 +967,7 @@ def user_check(name=None, uid=None):
 	else:
 		return None
 
-def user_ensure(name, passwd=None, home=None, uid=None, gid=None, shell=None):
+def user_ensure(name, passwd=None, home=None, uid=None, gid=None, shell=None, fullname=None, encrypted_passwd=True):
 	"""Ensures that the given users exists, optionally updating their
 	passwd/home/uid/gid/shell."""
 	d = user_check(name)
@@ -936,10 +983,12 @@ def user_ensure(name, passwd=None, home=None, uid=None, gid=None, shell=None):
 			options.append("-g '%s'" % (gid))
 		if shell != None and d.get("shell") != shell:
 			options.append("-s '%s'" % (shell))
+		if fullname != None and d.get("fullname") != fullname:
+			options.append("-c '%s'" % fullname)
 		if options:
 			sudo("usermod %s '%s'" % (" ".join(options), name))
 		if passwd:
-			user_passwd(name, passwd)
+			user_passwd(name=name, passwd=passwd, encrypted_passwd=encrypted_passwd)
 
 def user_remove(name, rmhome=None):
 	"""Removes the user with the given name, optionally
@@ -1014,7 +1063,11 @@ def group_user_del(group, user):
 				if group_for_user:
 						sudo("usermod -G '%s' '%s'" % (",".join(group_for_user), user))
 
-### ssh_<operation> functions
+# =============================================================================
+#
+# SSH
+#
+# =============================================================================
 
 def ssh_keygen(user, keytype="dsa"):
 	"""Generates a pair of ssh keys in the user's home .ssh directory."""
@@ -1032,12 +1085,6 @@ def ssh_keygen(user, keytype="dsa"):
 	else:
 		return key_file
 
-# =============================================================================
-#
-# MISC
-#
-# =============================================================================
-
 def ssh_authorize(user, key):
 	"""Adds the given key to the '.ssh/authorized_keys' for the given
 	user."""
@@ -1058,6 +1105,12 @@ def ssh_authorize(user, key):
 		file_write(keyf, key,             owner=user, group=user, mode="600")
 		return False
 
+# =============================================================================
+#
+# UPSTART
+#
+# =============================================================================
+
 def upstart_ensure(name):
 	"""Ensures that the given upstart service is running, restarting
 	it if necessary."""
@@ -1068,6 +1121,12 @@ def upstart_ensure(name):
 	else:
 		sudo("service %s restart" % name)
 
+# =============================================================================
+#
+# SYSTEM
+#
+# =============================================================================
+
 def system_uuid_alias_add():
 	"""Adds system UUID alias to /etc/hosts.
 	Some tools/processes rely/want the hostname as an alias in
@@ -1082,7 +1141,13 @@ def system_uuid():
 	"""Gets a machines UUID (Universally Unique Identifier)."""
 	return sudo('dmidecode -s system-uuid | tr "[A-Z]" "[a-z]"')
 
-#Only tested on Ubuntu!
+# =============================================================================
+#
+# LOCALE
+#
+# =============================================================================
+
+
 def locale_check(locale):
 	locale_data = sudo("locale -a | egrep '^%s$' ; true" % (locale,))
 	return locale_data == locale
